// hdstr.h contains the structure of the data on the hard drive 

#ifndef hdstr   
#define hdstr 1

#define int   short
#define uint  unsigned short
#define uchar unsigned char

#define ok 0
#define error -1
#define crash -2      
#define temp  -1

#define n_head 31
#define n_run  50
#define n_name 10

// the types of file
#define ft_erased    0
#define ft_ZXbasic   1
#define ft_ZXarray   2
#define ft_ZXarray$  3
#define ft_ZXcode    4
#define ft_ZXsnap48  5
#define ft_ZXmdrv    6
#define ft_ZXscreen  7
#define ft_ZXspecial 8
#define ft_ZXsnap128 9
#define ft_open      10
#define ft_ZXexecute 11
#define ft_UNIdir    12 // won't ever happen
#define ft_UNIcreate 13
//                   14
//                   15
#define ft_SAMbasic  16
#define ft_SAMarray  17
#define ft_SAMarray$ 18
#define ft_SAMcode   19
#define ft_SAMscreen 20
#define ft_SAMdir    21 // won't ever happen
//                   22
//                   23
//                   24
//                   25
//                   26
//                   27
#define ft_Hdos      28
#define ft_Hdir      29
#define ft_Hdisk     30
#define ft_Hfree     31
#define ft_Htmp      31  // temp files
#define ft_hidden    32
#define ft_protected 64
#define ft_archive   128

#define LABEL     0L
#define AVLSTART  1L
#define DIRSTART  2L
#define dirty 1
#define clean 0
#define removed -1
#define baddir  -2
#define namelen 10

#define HEOF         22
#define NOFILE       107
#define BADPATH      113
#define DISKFULL     105
#define EXISTS       109
#define BADNAME      18
#define INCONSISTANT 87
#define NOTEMPTY     114
#define TOOLONG      18
#define OUTOFMEMORY  1
#define TOOMANYAREAS 91
#define NUMTOOBIG    28
#define WRONGTYPE    94

//|S  

typedef struct s_FPCstack
{
  unsigned char reg_A,
                reg_B,
                reg_C,
                reg_D,
                reg_E,
                reg_H,
                reg_L;	
} s_FPCstack;

typedef struct s_samadr
{
 unsigned char page;   // page number 0 - 0x1F
 unsigned int  offset; // offset within page 0 - 0x3FFF
} s_samadr;

typedef struct s_fhead   // structure of a file header
{
 char  h_name[n_name];    // name of file
 short h_dir;             // number of directory 0 = root
 long  h_fsam;            // sector number of FSAM
} s_fhead;

typedef struct s_dir     // structure of a directory sector
{
 long    d_this;         // sector number of this directory entry
 long    d_next;         // sector number of next directory entry
 char    d_dirty;        // = 0 when same as on disk
 char    d_fill1[7];     // filler to get headers on boundary 
 s_fhead d_head[n_head]; // 31 file headers 
} s_dir;

typedef uchar s_buff[512];// one sector buffer

typedef struct s_chan	 // channel info
{
 long    c_filesize;	 // length of file in bytes
 long	 c_possn;	 // current possition in file
 long    c_cursect;      // address of current buffer n use
 uchar   c_dirty;        // dirty flag
 uchar   c_num;          // sam's channel number
 uint    c_offset;       // offset into buffer 0-511
} s_chan;


typedef struct s_time    // structure of a date&time entry
{
 unsigned char t_YY,     // year   0-99
      t_MM,              // month  1-12
      t_DD,              // day   1-31
      t_hh,              // hour   0-23
      t_mm,              // minute 0-59
      t_ss;              // second 0-59
} s_time;
//|S

typedef union s_info    // structure of file info
{
 char filler[50];       // a dummy entry to set the max size of the fib
 
 struct u_dirinfo       // directory fib
 {
  int dirnum;           // directory number of this directory
 } u_dirinfo;

 struct u_basic         // sam basic program fib
 {
  long  prog;        // length of program 
  long  vars;        // length of vars
  int      autorun;     // auto start line for basic
 } u_basic;

 struct u_code          // sam code file fib
 {
  long start;       // start mem address
  long length;      // length of code file
  long autorun;     // auto run address
 } u_code;

 struct u_screen        // sam screen file
 {
  char mode;            // screen mode 1-4
 } u_screen;

 struct u_diskimg       // image of a floppy disk
 {
  int tracks;           // number of tracks on this disk
  int dirtracks;        // number of directory tracks
 } u_diskimg;

 struct u_data          // data array
 {
  s_samadr length;      // length of array
 } u_data;
 
} s_info;      

typedef struct s_fib     // structure of a file info block
{
 unsigned char f_type;   // type of file 
 char    f_name[n_name]; // file name agian. useful for recovery
 s_time  f_cre;          // date & time of creation
 s_time  f_upd;          // date & time of last update
 s_time  f_acc;          // date & time of last access
 s_chan  f_chan;	 // channel info (only used in memory
 unsigned int f_count;   // counter inc every time file is acceessed
 s_info  f_info;         // file dependant info
} s_fib;
 

typedef struct s_fsam    // structure of a file sector allocation map
{
 long    f_myadr;        // sector address where this fsam goes
 long    f_next;         // sector address of next fsam for this file
 long    f_prev;         // sector address of prev fsam for this file
 char    f_dirty;        // 0 = same as on disk else needs to be written
 s_fib   f_fib;          // file info block
 uint    f_run;		 // current run number
 long    f_runst[n_run]; // start sector of a run
 long    f_runen[n_run]; // end sector of a run
} s_fsam;

typedef struct s_label   // structure of a SAM IDE disk label
{
 char    l_name[20];     // a user defined name for the drive
 char    l_sects;        // number of sectors 0-n 
 char    l_heads;        // number of heads   0-n
 int     l_tracks;       // number of tracks  0-nn
 s_time  l_cre;          // date & time of creation
 long    l_free;         // fsam of free space. always 1
 long    l_dir;          // first directory entry. always 2
 short   l_lastdir;      // number of last directory used 0-65500
} s_label;

typedef struct s_stream  // stream
{
	s_fsam	*fsam;	 // pointer to fsam 0 = closed
	char    *buff;	 // pointer to buffer
} s_stream;

#endif
