
#include "hdstr.h"
#include "hfile.h"
#include "hlow.h"
#include "hdir.h"
#include "hfname.h"
#include "hdio.h"
#include <string.h>
#include "hfree.h"
#include "hmalloc.h"

s_stream streams[16];
uint filesopen;

// **************************************************************************
// file handling routines
// **************************************************************************

int getfsam(char *path,s_fsam *myfsam )
// load the fsam into myfsam of a file given a pathname to that file
// return ok or error if not found
{
 long se;
 int dir;

 if ((dir = getdirnum(path)) == error)
 {
  hderror(BADPATH);
 }

 if ((se = finddirentry(getfilepart(path),dir)) == error)
 {
  hderror(NOFILE);
 }
 
 hdread(myfsam,se);
 return(ok);
}


int creat(char *path)
// create a file entry fail if already exists
// return ok or error
{
 long fsam;
 int dirnum;
 s_fhead *dirheader;
 char filepart[11];
 s_dir *mydir;
 s_fsam *myfsam;

 
 // test if exists
 if ((dirnum = getdirnum(path)) == error)
 {
  // bad path
  hderror(BADPATH);
 }
 
 strcpy(filepart,getfilepart(path)); // last 10 chars of path

 padname(filepart);
 
 if (filenametest(filepart) != ok)
 {
  // bad file name
  hderror(BADNAME);
 }

 if ((fsam = finddirentry(filepart,dirnum)) != error)
 {
  // file already exists
  hderror(EXISTS);
 }

 // got here so path is good and file is new
 fsam = getnextsector(); // space for new fsam

 if ((myfsam = calloc(1,sizeof(s_fsam))) == 0) hderror(OUTOFMEMORY);
 if ((mydir = calloc(1,sizeof(s_dir))) == 0) hderror(OUTOFMEMORY);

 // update directory
 dirheader = getnextfreedir(mydir); 
 memcpy(dirheader->h_name,filepart,10); 
 dirheader->h_dir=dirnum;
 dirheader->h_fsam=fsam;
 mydir->d_dirty=dirty;

 // update fsam
 memset(myfsam,0,sizeof(s_fsam));
 myfsam->f_myadr = fsam;
 gettime(&myfsam->f_fib.f_cre);
 gettime(&myfsam->f_fib.f_upd);
 gettime(&myfsam->f_fib.f_acc);
 memcpy(myfsam->f_fib.f_name,filepart,10);
 myfsam->f_fib.f_type = ft_Htmp; // until it is done proper
 myfsam->f_dirty=dirty;
 
 flushfsam(myfsam);
 flushdir(mydir);
 free(myfsam);
 free(mydir);
#if hdosdebug
 printf("\rcreat ok");
#endif
 return(ok);
}


int hopen(char *path, int channel)
{
// open a file for i/o 
// files are alwars opened i/o and must be created first
// returns error or a file handle 0-15 
   int handle;
   s_fsam *fsam;
   char  *buff;
   s_chan *chan;

   // if channel = temp then find highest free handle
   if (channel == temp)
   {
	   for (handle=15;(handle >= 0 && (streams[handle].fsam!=0);handle--);
	   if (handle < 0) return(error); // no handles left
   }

   if (channel != temp)
   { // test if already open
	   if (streams[channel].fsam != 0)
	   {
		   return (error) ;
	   }
	   handle = channel;
   }
   
   if ((fsam = calloc(1,sizeof(s_fsam)))==0) 
   {
	   hderror(OUTOFMEMORY);
   }

   if ((buff = calloc(1,sizeof(s_buff)))==0) 
   {
	   free(fsam);
	   hderror(OUTOFMEMORY);
   }

   if (getfsam(path,fsam)==error) 
   {
	   free(fsam);
	   free(buff);
	   return(error);
   }
   
   streams[handle].fsam = fsam; 
   streams[handle].buff = buff;

   
   fsam->f_fib.f_count++;
   gettime(&fsam->f_fib.f_acc);
   fsam->f_dirty=dirty;
   fsam->f_run = 0;
   chan = &fsam->f_fib.f_chan;
   chan->c_dirty=clean;
   chan->c_offset=0;
   chan->c_possn=0;
   

   if (fsam->f_runst[0]) 
   { // existing file so load first sector
   	hdread(buff,fsam->f_runst[0]);
   	chan->c_cursect=fsam->f_runst[0];
   }
   else 
   { // new file so clear buffer ready to write 
//	chan->c_cursect=0L; // will have been done by creat
//	chan->c_filesize=0; // will have been done by creat
	memset(buff,0,sizeof(s_buff));
   }
   filesopen++;
   return(handle);	// file is now open
}


void hflush(int handle)
{ // flush file buffers as needed
	s_fsam *fsam;

	fsam = streams[handle].fsam;
	if (!fsam) return; // wasn't open

	if (fsam->f_fib.f_chan.c_dirty==dirty)
	{
		fsam->f_fib.f_chan.c_dirty=clean;
		hdwrite(streams[handle].buff,
			fsam->f_fib.f_chan.c_cursect);
	}
	fsam->f_dirty=dirty;
 // 	flushfsam(fsam);	// only write when closing file
}

	
void hclose(int handle)
// close a file & update directory as required
// either does it or doesn't result is handle will close
{
	hflush(handle);
	if (streams[handle].fsam) flushfsam(streams[handle].fsam);
	free(streams[handle].fsam);
	streams[handle].fsam=0;
	free(streams[handle].buff);
	streams[handle].buff=0;
	filesopen--;
	if (myavl) 
	{
		flushfsam(myavl);
		if (!filesopen)
			
		{	
			free(myavl);
			myavl=0;
		}
	}	
}		

int getnextfilesector(s_fsam* fsam, s_chan* chan,int handle, char rw)
// get the next sector for a file if rw = r(ead) then return error if
// past EOF; if rw = w(rite) then return error if unable to extend file
// if ok then fsam & buff are ready for next read or write operation
{
	long cursect;

	hflush(handle);
	
	if ((fsam = streams[handle].fsam) == 0)
		 return(error); // not open

	chan = &fsam->f_fib.f_chan;

	if (!chan->c_cursect) // no current sector yet
	{
		if (!fsam->f_runst[0]) // never been a sector 
		if (rw == 'r') return(error); // eof on read
		else // writing
		{
		   chan->c_cursect = getnextsector();
		   chan->c_offset = 0;
		   fsam->f_runst[0]=chan->c_cursect;
		   fsam->f_runen[0]=chan->c_cursect;	
		   fsam->f_run=0;
		   memset(streams[handle].buff,0,sizeof(s_buff));
		   return(ok);
		}
	}
	
	// just get next sect in this run
	if (++chan->c_cursect <= fsam->f_runen[fsam->f_run])
	{
		hdread(streams[handle].buff,chan->c_cursect);
		chan->c_offset=0;
		return(ok);
	}

	cursect = chan->c_cursect; // ie next contigious sector


 	// test if all runs in this fsam have been used
 	if (++fsam->f_run < n_run)
 	{
 		// next sector = start of next run
 		chan->c_cursect = fsam->f_runst[fsam->f_run];

		// will be 0 if this run has not been used
		if (!chan->c_cursect) 
		{  
			// test if at end of file
			if (rw == 'r') 	return(error);

			// if writing then extend file by 1 sector
			chan->c_cursect = getnextsector();

			// if next sect is not = 1+last sector then
			// set up next run else inc prev runend
			if (cursect != chan->c_cursect)
			{
				fsam->f_runst[fsam->f_run] = chan->c_cursect;
				fsam->f_runen[fsam->f_run] = chan->c_cursect;
			}
			else
			{
				fsam->f_run--;
				fsam->f_runen[fsam->f_run]++;
			}
			fsam->f_dirty=dirty;
		}
		hdread(streams[handle].buff,chan->c_cursect);
		chan->c_offset=0;
		return(ok);
	}

	hderror(TOOMANYAREAS);
	return(0);	// code will not get here
}


int hgetc(int handle)
// get next char from file return -1 if at end of file or not open
{
	s_fsam *fsam;
	s_chan *chan;

	if ((fsam = streams[handle].fsam) == 0)
	{ // file not open
		hderror(47); 
	}
	
	chan = &fsam->f_fib.f_chan;

	if (chan->c_possn >= chan->c_filesize) 
	{ // EOF
		hderror(22);
	}
	
	if (chan->c_offset < sizeof(s_buff))
	{ // just return next char
		chan->c_possn++;
		return(0x00ff 
		& streams[handle].buff[chan->c_offset++]);
	}
		
	// need to get next sector
	if (getnextfilesector(fsam,chan,handle,'r')==error)
	{
		return(error);
	}
	
	return(hgetc(handle);
	
}

int hputc(char ch,int handle)
// write char c to file handle return ok or error
{
	s_fsam *fsam;
	s_chan *chan;

	
	if ((fsam = streams[handle].fsam) == 0)
	{ // not open
		hderror(47); 
	}
	
	chan = &fsam->f_fib.f_chan;
	
	if (chan->c_offset >= sizeof(s_buff))
	{
		if ((getnextfilesector(fsam,chan,handle,'w')) == error)
		return(error);
		// else
		return(hputc(ch,handle));
		
	}
	if (!chan->c_possn) // first write to this file
	{
		if ((getnextfilesector(fsam,chan,handle,'w')) == error)
		return(error);
	}
	
	streams[handle].buff[chan->c_offset] = ch;
	chan->c_offset++;
	chan->c_dirty=dirty;
	chan->c_possn++;
	
	if (chan->c_possn > chan->c_filesize)
	{
		chan->c_filesize++;
	}
	return(ok);
}


// **************************************************************************
// end of file handling routines
// **************************************************************************
